from rest_framework.schemas.openapi import (
    SchemaGenerator as DRFSchemaGenerator,
    AutoSchema as DRFAutoSchema,
)
from camomilla.contrib.rest_framework.serializer import (
    TranslationsMixin,
    plain_to_nest,
    TRANS_ACCESSOR,
)
from camomilla.serializers.fields.json import StructuredJSONField
from camomilla.utils.getters import find_and_replace_dict


class AutoSchema(DRFAutoSchema):
    extra_components = {}

    def map_serializer(self, serializer):
        schema = super(AutoSchema, self).map_serializer(serializer)
        if isinstance(serializer, TranslationsMixin) and serializer.is_translatable:
            schema = plain_to_nest(schema["properties"], serializer.translation_fields)
            schema[TRANS_ACCESSOR] = {
                "type": "object",
                "properties": {
                    k: {"type": "object", "properties": v}
                    for k, v in schema[TRANS_ACCESSOR].items()
                },
            }
        return schema

    def get_components(self, path, method):
        components = super().get_components(path, method)
        if len(self.extra_components.keys()) > 0:
            components = {**(components or {}), **self.extra_components}
        return components

    def map_field(self, field):
        if isinstance(field, StructuredJSONField):

            def replace(key, value):
                if isinstance(value, str) and value.startswith("#/definitions"):
                    return value.replace("#/definitions", "#/components/schemas")
                return value

            self.extra_components.update(
                **find_and_replace_dict(
                    field.json_schema.pop("definitions", {}), replace
                )
            )

            return find_and_replace_dict(field.json_schema, replace)
        return super().map_field(field)


class SchemaGenerator(DRFSchemaGenerator):
    def create_view(self, callback, method, request=None):
        view = super(SchemaGenerator, self).create_view(callback, method, request)
        view.schema = AutoSchema()
        return view



from django.db import models
from rest_framework import serializers

from camomilla import structured

from .json import StructuredJSONField
from .file import FileField, ImageField
from .related import RelatedField


class FieldsOverrideMixin:
    """
    This mixin automatically overrides the fields of the serializer with camomilla's backed ones.
    """
    serializer_field_mapping = {
        **serializers.ModelSerializer.serializer_field_mapping,
        models.FileField: FileField,
        models.ImageField: ImageField,
        structured.StructuredJSONField: StructuredJSONField,
    }
    serializer_related_field = RelatedField



from ckeditor_uploader.widgets import CKEditorUploadingWidget
from django import forms
from django.contrib import admin
from django.http import HttpResponse

from camomilla import settings

if settings.ENABLE_TRANSLATIONS:
    from modeltranslation.admin import (
        TabbedTranslationAdmin as TranslationAwareModelAdmin,
    )
else:
    from django.contrib.admin import ModelAdmin as TranslationAwareModelAdmin

from camomilla.models import Article, Content, Media, MediaFolder, Page, Tag, Menu


class AbstractPageAdmin(TranslationAwareModelAdmin):
    change_form_template = "admin/camomilla/page/change_form.html"


class UserProfileAdmin(admin.ModelAdmin):
    pass


class ArticleAdminForm(forms.ModelForm):
    class Meta:
        model = Article
        exclude = ("slug",)
        widgets = {"content": CKEditorUploadingWidget}


class ArticleAdmin(AbstractPageAdmin):
    filter_horizontal = ("tags",)
    form = ArticleAdminForm


class TagAdmin(TranslationAwareModelAdmin):
    pass


class MediaFolderAdmin(admin.ModelAdmin):
    readonly_fields = ("path",)


class ContentAdminForm(forms.ModelForm):
    class Meta:
        model = Content
        fields = "__all__"
        widgets = {"content": CKEditorUploadingWidget}


class ContentAdmin(TranslationAwareModelAdmin):
    form = ContentAdminForm


class MediaAdmin(TranslationAwareModelAdmin):
    exclude = (
        "thumbnail",
        "size",
        "image_props",
    )
    readonly_fields = ("image_preview", "image_thumb_preview", "mime_type")
    list_display = (
        "__str__",
        "title",
        "image_thumb_preview",
    )

    def response_add(self, request, obj):
        if request.GET.get("_popup", ""):
            return HttpResponse(
                """
               <script type="text/javascript">
                  opener.dismissAddRelatedObjectPopup(window, %s, '%s');
               </script>"""
                % (obj.id, obj.json_repr)
            )
        else:
            return super(MediaAdmin, self).response_add(request, obj)


class PageAdmin(AbstractPageAdmin):
    readonly_fields = ("permalink",)


class MenuAdmin(TranslationAwareModelAdmin):
    pass


admin.site.register(Article, ArticleAdmin)
admin.site.register(MediaFolder, MediaFolderAdmin)
admin.site.register(Tag, TagAdmin)
admin.site.register(Content, ContentAdmin)
admin.site.register(Media, MediaAdmin)
admin.site.register(Page, PageAdmin)
admin.site.register(Menu, MenuAdmin)



from django.contrib.contenttypes.models import ContentType
from rest_framework import serializers


class ContentTypeSerializer(serializers.ModelSerializer):
    verbose_name = serializers.SerializerMethodField()
    verbose_name_plural = serializers.SerializerMethodField()

    def get_verbose_name(self, obj):
        return obj.model_class()._meta.verbose_name.title()

    def get_verbose_name_plural(self, obj):
        return obj.model_class()._meta.verbose_name_plural.title()

    class Meta:
        model = ContentType
        fields = "__all__"



def get_standard_bases() -> tuple:
    from rest_framework.serializers import ModelSerializer
    from camomilla.serializers.fields import FieldsOverrideMixin
    from camomilla.contrib.rest_framework.serializer import RemoveTranslationsMixin
    from camomilla.serializers.mixins import (
        JSONFieldPatchMixin,
        NestMixin,
        OrderingMixin,
        SetupEagerLoadingMixin,
    )

    return (
        NestMixin,
        FieldsOverrideMixin,
        JSONFieldPatchMixin,
        OrderingMixin,
        RemoveTranslationsMixin,
        SetupEagerLoadingMixin,
        ModelSerializer,
    )


def build_standard_model_serializer(model, depth, bases=None):
    if bases is None:
        bases = get_standard_bases()
    return type(
        f"{model.__name__}StandardSerializer",
        bases,
        {
            "Meta": type(
                "Meta",
                (object,),
                {"model": model, "depth": depth, "fields": "__all__"},
            )
        },
    )



from collections import OrderedDict
from rest_framework import serializers, relations


class RelatedField(serializers.PrimaryKeyRelatedField):
    """
    This field helps to serialize and deserialize related data.
    It serializes related data building a nested serializer.
    Allowing insertions with both nested and plain data.


    For example it accepts as input data both:
    ```json
    {"related_field": 1}
    {"related_field": {"id": 1, "field": "value"}}
    ```

    """

    def __init__(self, **kwargs):
        self.serializer = kwargs.pop("serializer", None)
        self.lookup = kwargs.pop("lookup", "id")
        if self.serializer is not None:
            assert issubclass(
                self.serializer, serializers.Serializer
            ), '"serializer" is not a valid serializer class'
            assert hasattr(
                self.serializer.Meta, "model"
            ), 'Class {serializer_class} missing "Meta.model" attribute'.format(
                serializer_class=self.serializer.__class__.__name__
            )
            if not kwargs.get("read_only", False):
                kwargs["queryset"] = kwargs.get(
                    "queryset", self.serializer.Meta.model.objects.all()
                )
            self.allow_insert = kwargs.pop("allow_insert", False)
            # kwargs["allow_null"] = kwargs.get("allow_null", self.serializer.Meta.model._meta.get_field(self.source).null)
        super().__init__(**kwargs)

    def use_pk_only_optimization(self):
        return False if self.serializer else True

    def to_representation(self, instance):
        if self.serializer:
            return self.serializer(instance, context=self.context).data
        return super().to_representation(instance)

    def to_internal_value(self, data):
        if isinstance(data, dict):
            instance = (
                self.get_queryset()
                .filter(**{self.lookup: data.get(self.lookup, None)})
                .first()
            )
            if self.allow_insert is True and len(data.keys()) and self.serializer:
                serialized_data = self.serializer(
                    instance=instance, data=data, context=self.context
                )
                if serialized_data.is_valid(raise_exception=ValueError):
                    instance = serialized_data.save()
            return instance
        return super().to_internal_value(data)

    def get_choices(self, cutoff=None):
        queryset = self.get_queryset()
        if queryset is None:
            return {}
        if cutoff is not None:
            queryset = queryset[:cutoff]

        return OrderedDict(
            [
                (
                    super(RelatedField, self).to_representation(item),
                    self.display_value(item),
                )
                for item in queryset
            ]
        )

    @classmethod
    def many_init(cls, *args, **kwargs):
        class ManyRelatedField(serializers.ManyRelatedField):
            def to_internal_value(self, data):
                if isinstance(data, str) or not hasattr(data, "__iter__"):
                    self.fail("not_a_list", input_type=type(data).__name__)
                if not self.allow_empty and len(data) == 0:
                    self.fail("empty")

                child = self.child_relation
                instances = {
                    getattr(item, child.lookup): item
                    for item in child.get_queryset().filter(
                        **{
                            f"{child.lookup}__in": [
                                item.get(child.lookup, None)
                                if isinstance(item, dict)
                                else item
                                for item in data
                            ]
                        }
                    )
                }
                if child.allow_insert is True and len(data.keys()) and child.serializer:
                    for item in data:
                        serialized_data = child.serializer(
                            instance=instances.get(item.get(child.lookup)),
                            data=item,
                            context=child.context,
                        )
                        if serialized_data.is_valid(raise_exception=ValueError):
                            instance = serialized_data.save()
                            instances[instance.pk] = instance
                return instances.values()

        list_kwargs = {"child_relation": cls(*args, **kwargs)}
        for key in kwargs:
            if key in relations.MANY_RELATION_KWARGS:
                list_kwargs[key] = kwargs[key]
        return ManyRelatedField(**list_kwargs)



import os

from django.utils.translation import gettext_lazy as _

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

STATIC_ROOT = os.path.join(BASE_DIR, "static")

STATICFILES_DIRS = []
if os.path.exists(os.path.join(BASE_DIR, "build")):
    STATICFILES_DIRS.append(os.path.join(BASE_DIR, "build"))
if os.path.exists(os.path.join(BASE_DIR, "public")):
    STATICFILES_DIRS.append(os.path.join(BASE_DIR, "public"))

MEDIA_ROOT = os.path.join(BASE_DIR, "media")
MEDIA_URL = "/media/"

THUMB_FOLDER = "thumbnails"

REST_FRAMEWORK = {
    "DEFAULT_PARSER_CLASSES": (
        "rest_framework.parsers.JSONParser",
        "rest_framework.parsers.FormParser",
        "rest_framework.parsers.MultiPartParser",
    ),
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.TokenAuthentication"
        "camomilla.authentication.SessionAuthentication"
    ),
}

ADMIN_SITE_HEADER = _("Camomilla advanced panel")

CAMOMILLA_THUMBNAIL_WIDTH = 50
CAMOMILLA_THUMBNAIL_HEIGHT = 50


PNG_OPTIMIZATION_COMMAND = "pngquant {0} -f --ext .png"
JPEG_OPTIMIZATION_COMMAND = "jpegoptim {0}"

LANG_ON_PREFERENCE_DISABLED_VIEWS = ["sitemap", "sitemap_xml"]

HVAD = {"AUTOLOAD_TRANSLATIONS": True}



from collections import defaultdict
from inspect import isclass
from typing import Any, Dict, Sequence, Type, Iterable, Union
from typing_extensions import get_args, get_origin
from camomilla.settings import STRUCTURED_FIELD_CACHE_ENABLED
from camomilla.structured.fields import ForeignKey, QuerySet
from camomilla.structured.models import BaseModel
from camomilla.structured.utils import _LazyType, get_type, pointed_setter
from django.db.models import Model as DjangoModel
from camomilla.utils.getters import pointed_getter


# TODO:
# ::: Actually this is a first draft.
# The system should be more type safe.
# It should handle initialization made with model instance not only dicts
# Neet to check if there are problems with different formats for pks (model instaces or string or dicts)


class Cache(dict):
    pass


class ValueWithCache:
    def __init__(self, cache, model, value) -> None:
        self.cache: Cache = cache
        self.value: Union[Iterable[Union[str, int]], str, int] = value
        self.model: Type[DjangoModel] = model

    def retrieve(self):
        cache = self.cache.get(self.model)
        if hasattr(self.value, '__iter__'):
            qs = self.model.objects.filter(pk__in=self.value)
            setattr(
                qs,
                "_result_cache",
                [cache[i] for i in self.value if i in cache] if cache else [],
            )
            return qs
        else:
            val = cache.get(self.value, None)
            if val is None:
                return self.model._default_manager.get(pk=self.value)
            else:
                return val


class RelInfo:
    FKField: str = "fk"
    QSField: str = "qs"
    RIField: str = "rel"
    RLField: str = "rel_l"

    def __init__(self, model, type, field) -> None:
        self.model = model
        self.type = type
        self.field = field


class CacheBuilder:
    def __init__(self, related_fields: Dict[str, RelInfo]) -> None:
        self.__related_fields__ = related_fields

    @classmethod
    def from_model(cls, model: BaseModel) -> "CacheBuilder":
        return cls(related_fields=cls.inspect_related_fields(model))

    @classmethod
    def inspect_related_fields(cls, model: Type[BaseModel]) -> Dict[str, RelInfo]:
        related = {}
        for field_name, field in model.model_fields.items():
            annotation = field.annotation
            origin = get_origin(annotation)
            args = get_args(annotation)

            if isclass(origin) and issubclass(origin, ForeignKey):
                related[field_name] = RelInfo(
                    get_type(annotation), RelInfo.FKField, field
                )
            elif isclass(origin) and issubclass(origin, QuerySet):
                related[field_name] = RelInfo(
                    get_type(annotation), RelInfo.QSField, field
                )

            elif isclass(origin) and issubclass(origin, Sequence):
                lazy_types = [
                    _LazyType(arg).evaluate(model)
                    for arg in args
                    if isinstance(arg, str)
                ]
                subclass = next(
                    (
                        c
                        for c in list(args) + lazy_types
                        if isclass(c) and issubclass(c, BaseModel)
                    ),
                    None,
                )
                if isclass(subclass) and issubclass(subclass, BaseModel):
                    related[field_name] = RelInfo(subclass, RelInfo.RLField, field)

            elif isclass(annotation) and issubclass(annotation, BaseModel):
                related[field_name] = RelInfo(annotation, RelInfo.RIField, field)
        return related

    def get_all_fk_data(self, data):
        if isinstance(data, Sequence):
            fk_data = defaultdict(list)
            for index in range(len(data)):
                child_fk_data = self.get_all_fk_data(data[index])
                for model, touples in child_fk_data.items():
                    fk_data[model] += [(f"{index}.{t[0]}", t[1]) for t in touples]
            return fk_data
        return self.get_fk_data(data)

    def get_fk_data(self, data):
        fk_data = defaultdict(list)
        if not data:
            return fk_data
        for field_name, info in self.__related_fields__.items():
            if info.type == RelInfo.FKField:
                value = pointed_getter(data, field_name, None)
                if value:
                    if isinstance(
                        value, ValueWithCache
                    ):  # needed to break recursive cache builds
                        fk_data = {}
                        break
                    fk_data[info.model].append(
                        (
                            field_name,
                            pointed_getter(value, info.model._meta.pk.attname, value),
                        )
                    )
            elif info.type == RelInfo.QSField:
                value = pointed_getter(data, field_name, [])
                if isinstance(value, list):
                    if any(
                        True for v in value if isinstance(v, ValueWithCache)
                    ):  # needed to break recursive cache builds
                        fk_data = {}
                        break
                    fk_data[info.model].append(
                        (
                            field_name,
                            [
                                pointed_getter(i, info.model._meta.pk.attname, i)
                                for i in value
                                if i
                            ],
                        )
                    )

            elif info.type == RelInfo.RLField:
                values = pointed_getter(data, field_name, [])
                if isinstance(values, list):
                    for index in range(len(values)):
                        for model, touples in (
                            self.from_model(info.model)
                            .get_all_fk_data(values[index])
                            .items()
                        ):
                            fk_data[model] += [
                                (f"{field_name}.{index}.{t[0]}", t[1]) for t in touples
                            ]
            elif info.type == RelInfo.RIField:
                value = pointed_getter(data, field_name, None)
                child_fk_data = self.from_model(info.model).get_all_fk_data(value)
                for model, touples in child_fk_data.items():
                    fk_data[model] += [(f"{field_name}.{t[0]}", t[1]) for t in touples]
        return fk_data

    def inject_cache(self, data: Any) -> Any:
        if not STRUCTURED_FIELD_CACHE_ENABLED:
            return data
        fk_data = self.get_all_fk_data(data)
        plainset = defaultdict(set)
        for model, touples in fk_data.items():
            for t in touples:
                if isinstance(t[1], Sequence):
                    plainset[model].update(t[1])
                else:
                    plainset[model].add(t[1])
        cache = Cache()
        for model, values in plainset.items():
            models = []
            pks = []
            for value in values:
                if isinstance(value, model):
                    models.append(value)
                else:
                    pks.append(value)
            models_pks = [m.pk for m in models]
            pks = [pk for pk in pks if pk not in models_pks]
            if len(pks):
                models += list(model.objects.filter(pk__in=pks))
            cache[model] = {obj.pk: obj for obj in models}

        for model, touples in fk_data.items():
            for t in touples:
                pointed_setter(data, t[0], ValueWithCache(cache, model, t[1]))
        return data



from django.forms import ValidationError
from camomilla.serializers.base import BaseModelSerializer
from django.contrib.auth import get_user_model, password_validation
from django.contrib.auth.models import Permission
from rest_framework import serializers
from django.utils.translation import gettext_lazy as _
from django.db.models import Q


class PermissionSerializer(BaseModelSerializer):
    class Meta:
        model = Permission
        fields = "__all__"


class UserProfileSerializer(BaseModelSerializer):
    user_permissions = PermissionSerializer(read_only=True, many=True)
    group_permissions = serializers.SerializerMethodField()
    all_permissions = serializers.SerializerMethodField()
    password = serializers.CharField(
        write_only=True, required=False, allow_null=True, allow_blank=True
    )
    repassword = serializers.CharField(
        write_only=True, required=False, allow_null=True, allow_blank=True
    )

    class Meta:
        model = get_user_model()
        fields = "__all__"

    def get_group_permissions(self, instance):
        return PermissionSerializer(
            Permission.objects.filter(
                group__pk__in=instance.groups.values_list("pk", flat=True)
            ),
            context=self.context,
            many=True,
        ).data

    def get_all_permissions(self, instance):
        return PermissionSerializer(
            Permission.objects.filter(
                Q(group__pk__in=instance.groups.values_list("pk", flat=True))
                | Q(pk__in=instance.user_permissions.values_list("pk", flat=True))
            ),
            context=self.context,
            many=True,
        ).data

    def validate_password(self, value):
        password_validation.validate_password(value)
        return value

    def validate_repassword(self, value):
        if "password" in self.initial_data:
            if not value:
                raise ValidationError(_("This field is required"))
            if self.initial_data["password"] != value:
                raise ValidationError(_("Passwords do not match."))
        return value

    def update(self, instance, validated_data):
        user = super(UserSerializer, self).update(instance, validated_data)
        if "password" in validated_data:
            user.set_password(validated_data["password"])
            user.save()
        return user


class UserSerializer(BaseModelSerializer):
    id = serializers.IntegerField(read_only=True)
    password = serializers.CharField(
        write_only=True, required=False, allow_null=True, allow_blank=True
    )
    level = serializers.CharField(required=False)
    has_token = serializers.SerializerMethodField("get_token", read_only=True)

    def get_token(self, obj):
        return hasattr(obj, "auth_token")

    class Meta:
        model = get_user_model()
        fields = (
            "id",
            "username",
            "first_name",
            "last_name",
            "email",
            "password",
            "level",
            "user_permissions",
            "has_token",
            "is_superuser",
        )

    def validate_password(self, value):
        password_validation.validate_password(value)
        return value

    def create(self, validated_data):
        user = super(UserSerializer, self).create(validated_data)
        user.set_password(validated_data["password"])
        user.save()
        return user

    def update(self, instance, validated_data):
        user = super(UserSerializer, self).update(instance, validated_data)
        if "password" in validated_data:
            user.set_password(validated_data["password"])
            user.save()
        return user



from django.conf import settings as django_settings
from modeltranslation.settings import ENABLE_REGISTRATIONS

from camomilla.utils.getters import pointed_getter

PROJECT_TITLE = pointed_getter(
    django_settings,
    "CAMOMILLA.PROJECT_TITLE",
    pointed_getter(django_settings, "CAMOMILLA_PROJECT_TITLE", "Camomilla"),
)

THUMBNAIL_FOLDER = pointed_getter(
    django_settings,
    "CAMOMILLA.MEDIA.THUMBNAIL.FOLDER",
    pointed_getter(django_settings, "CAMOMILLA_THUMBTHUMBNAIL_FOLDER", "thumbnails"),
)
THUMBNAIL_WIDTH = pointed_getter(
    django_settings,
    "CAMOMILLA.MEDIA.THUMBNAIL.WIDTH",
    pointed_getter(django_settings, "CAMOMILLA_THUMBNAIL_WIDTH", 50),
)
THUMBNAIL_HEIGHT = pointed_getter(
    django_settings,
    "CAMOMILLA.MEDIA.THUMBNAIL.HEIGHT",
    pointed_getter(django_settings, "CAMOMILLA_THUMBNAIL_HEIGHT", 50),
)
BASE_URL = pointed_getter(
    django_settings,
    "CAMOMILLA.ROUTER.BASE_URL",
    pointed_getter(django_settings, "FORCE_SCRIPT_NAME", None),
)
BASE_URL = BASE_URL and "/" + BASE_URL.strip("/")


ARTICLE_DEFAULT_TEMPLATE = pointed_getter(
    django_settings,
    "CAMOMILLA.RENDER.ARTICLE.DEFAULT_TEMPLATE",
    "defaults/articles/default.html",
)
PAGE_DEFAULT_TEMPLATE = pointed_getter(
    django_settings,
    "CAMOMILLA.RENDER.PAGE.DEFAULT_TEMPLATE",
    "defaults/pages/default.html",
)
ARTICLE_INJECT_CONTEXT_FUNC = pointed_getter(
    django_settings, "CAMOMILLA.RENDER.ARTICLE.INJECT_CONTEXT", None
)
PAGE_INJECT_CONTEXT_FUNC = pointed_getter(
    django_settings, "CAMOMILLA.RENDER.PAGE.INJECT_CONTEXT", None
)

ENABLE_TRANSLATIONS = (
    ENABLE_REGISTRATIONS and "modeltranslation" in django_settings.INSTALLED_APPS
)

MEDIA_OPTIMIZE_MAX_WIDTH = pointed_getter(
    django_settings, "CAMOMILLA.MEDIA.OPTIMIZE.MAX_WIDTH", 1980
)
MEDIA_OPTIMIZE_MAX_HEIGHT = pointed_getter(
    django_settings, "CAMOMILLA.MEDIA.OPTIMIZE.MAX_HEIGHT", 1400
)
MEDIA_OPTIMIZE_DPI = pointed_getter(django_settings, "CAMOMILLA.MEDIA.OPTIMIZE.DPI", 30)

MEDIA_OPTIMIZE_JPEG_QUALITY = pointed_getter(
    django_settings, "CAMOMILLA.MEDIA.OPTIMIZE.JPEG_QUALITY", 85
)

ENABLE_MEDIA_OPTIMIZATION = pointed_getter(
    django_settings, "CAMOMILLA.MEDIA.OPTIMIZE.ENABLE", True
)

API_NESTING_DEPTH = pointed_getter(django_settings, "CAMOMILLA.API.NESTING_DEPTH", 10)

AUTO_CREATE_HOMEPAGE = pointed_getter(
    django_settings, "CAMOMILLA.RENDER.AUTO_CREATE_HOMEPAGE", True
)

TEMPLATE_CONTEXT_FILES = pointed_getter(
    django_settings, "CAMOMILLA.RENDER.TEMPLATE_CONTEXT_FILES", []
)

STRUCTURED_FIELD_CACHE_ENABLED = pointed_getter(
    django_settings, "CAMOMILLA.STRUCTURED_FIELD.CACHE_ENABLED", True
)

DEBUG = pointed_getter(django_settings, "CAMOMILLA.DEBUG", django_settings.DEBUG)

# camomilla settings example
# CAMOMILLA = {
#     "PROJECT_TITLE": "",
#     "ROUTER": {
#         "BASE_URL": ""
#     },
#     "MEDIA": {
#         "OPTIMIZE": {"MAX_WIDTH": 1980, "MAX_HEIGHT": 1400, "DPI": 30, "JPEG_QUALITY": 85, "ENABLE": True},
#         "THUMBNAIL": {"FOLDER": "", "WIDTH": 50, "HEIGHT": 50}
#     },
#     "RENDER": {
#         "TEMPLATE_CONTEXT_FILES": [],
#         "AUTO_CREATE_HOMEPAGE": True,
#         "ARTICLE": {"DEFAULT_TEMPLATE": "", "INJECT_CONTEXT": None },
#         "PAGE": {"DEFAULT_TEMPLATE": "", "INJECT_CONTEXT": None }
#     },
#     "STRUCTURED_FIELD": {
#         "CACHE_ENABLED": True
#     }
#     "API": {"NESTING_DEPTH": 10 },
#     "DEBUG": False
# }



import os
import mock

from camomilla.utils import get_host_url, get_complete_url, get_page
from camomilla.exceptions import NeedARedirect
from camomilla.models import Page, Article

from django.http import Http404
from django.test import TestCase
from django.test import RequestFactory
from django.utils.translation import activate, get_language

from requests import RequestException


class UtilsTestCase(TestCase):
    def setUp(self):
        pass

    def test_get_host_url(self):
        """Our beloved get_host_url utility"""
        request_factory = RequestFactory()
        request = request_factory.get("/path")
        request.META["HTTP_HOST"] = "localhost"
        host_url = get_host_url(request)
        self.assertEqual(host_url, "http://localhost")
        host_url = get_host_url(None)
        self.assertEqual(host_url, None)

    def test_get_complete_url(self):
        """Our beloved get_complete_url utility"""
        request_factory = RequestFactory()
        request = request_factory.get("/path")
        request.META["HTTP_HOST"] = "localhost"
        complete_url = get_complete_url(request, "path")
        self.assertEqual(complete_url, "http://localhost/path")
        complete_url = get_complete_url(request, "path", "it")
        self.assertEqual(complete_url, "http://localhost/path")
        complete_url = get_complete_url(request, "path", "fr")
        self.assertEqual(complete_url, "http://localhost/fr/path")

    def test_get_page_with_default_seo(self):
        """Our beloved get_seo utility with auto attributes"""
        request_factory = RequestFactory()
        request = request_factory.get("/path")
        request.META["HTTP_HOST"] = "localhost"
        page = Page.get(request, identifier="home")
        self.assertEqual(page.og_url, "http://localhost/path")
        self.assertEqual(page.canonical, "http://localhost/path")

    def test_get_article_with_default_seo(self):
        """Our beloved get_seo utility with auto attributes"""
        request_factory = RequestFactory()
        request = request_factory.get("/path")
        request.META["HTTP_HOST"] = "localhost"
        Article.objects.create(permalink="main")
        article = Article.get(request, permalink="main")
        self.assertEqual(article.og_url, "http://localhost/path")
        self.assertEqual(article.canonical, "http://localhost/path")

    def test_compile_seo_overwrite(self):
        """Our beloved get_seo utility with auto attributes"""
        request_factory = RequestFactory()
        request = request_factory.get("/path")
        request.META["HTTP_HOST"] = "localhost"
        article = Article.objects.create(permalink="main")
        article.canonical = "/myarticle"
        article.og_url = "/myarticle"
        article.save()
        article = Article.get(request, permalink="main")
        self.assertEqual(article.og_url, "http://localhost/myarticle")
        self.assertEqual(article.canonical, "http://localhost/myarticle")

    def test_get_article_with_redirect(self):
        """Our beloved get_seo utility with auto attributes"""
        request_factory = RequestFactory()
        request = request_factory.get("/articles/articolo-1")
        request.META["HTTP_HOST"] = "localhost"

        article = Article.objects.create(permalink="article-1", language_code="en")
        article.translate("it")
        article.permalink = "articolo-1"
        article.save()
        self.assertRaises(NeedARedirect, Article.get, request, permalink="article-1")
        activate("de")
        self.assertRaises(Http404, Article.get, request, permalink="article-1")



from camomilla.models import Article
from camomilla.permissions import CamomillaBasePermissions
from camomilla.serializers import ArticleSerializer
from camomilla.views.base import BaseModelViewset
from camomilla.views.mixins import BulkDeleteMixin, GetUserLanguageMixin


class ArticleViewSet(GetUserLanguageMixin, BulkDeleteMixin, BaseModelViewset):
    queryset = Article.objects.all()
    serializer_class = ArticleSerializer
    permission_classes = (CamomillaBasePermissions,)
    search_fields = ["title", "identifier", "content", "permalink"]
    model = Article



import pytest
from django.contrib.auth.models import User
from rest_framework.test import APIClient

from camomilla.models import Tag

client = APIClient()


def login_superuser():
    User.objects.create_superuser("admin", "myemail@test.com", "adminadmin")
    response = client.post(
        "/api/camomilla/token-auth/", {"username": "admin", "password": "adminadmin"}
    )
    return response.json()["token"]


@pytest.mark.django_db
def test_create_tag_no_access():
    response = client.post("/api/camomilla/tags/", {"name_en": "First tag"})
    assert response.status_code == 401


@pytest.mark.django_db
def test_crud_tag():
    # Create
    token = login_superuser()
    client.credentials(HTTP_AUTHORIZATION="Token " + token)
    response = client.post("/api/camomilla/tags/", {"name_en": "First tag"})
    
    assert response.json()["name"] == "First tag"
    assert len(Tag.objects.all()) == 1
    assert response.status_code == 201

    # Create another with a different language
    response = client.post("/api/camomilla/tags/", {"name_it": "Secondo tag"})
    assert response.json()["translations"]["it"]["name"] == "Secondo tag"
    assert len(Tag.objects.all()) == 2
    assert response.status_code == 201

    # Translate the second one in english
    response = client.patch(
        "/api/camomilla/tags/2/",
        {"translations": {"en": {"name": "Second tag"}}},
        format="json",
    )
    assert response.json()["translations"]["en"]["name"] == "Second tag"
    assert response.json()["translations"]["it"]["name"] == "Secondo tag"
    assert len(Tag.objects.all()) == 2

    assert response.status_code == 200

    # Get the tags in english
    response = client.get("/api/camomilla/tags/")

    assert response.json()[0]["name"] == "Second tag"

    # Get the tags in italianith fallbacks!
    response = client.get("/api/camomilla/tags/?language=it")

    assert response.json()[0]["name"] == "Secondo tag"

    # Delete the tag
    response = client.delete("/api/camomilla/tags/2/")

    assert len(Tag.objects.all()) == 1

    assert response.status_code == 204



from collections import defaultdict
from typing import Callable, Optional, TYPE_CHECKING, Type, Dict
import inspect
from django.http import HttpRequest

if TYPE_CHECKING:
    from camomilla.models.page import AbstractPage


class PagesContextRegistry:
    def __init__(self) -> None:
        self._model_registry = defaultdict(set)
        self._template_registry = defaultdict(set)

    def register(
        self,
        func: Callable,
        template_path: Optional[str],
        page_model: Optional[Type["AbstractPage"]],
    ):
        assert (
            template_path is not None or page_model is not None
        ), "You must provide at least one between template_path and page_model"
        if template_path is not None:
            self._template_registry[template_path].add(func)
        if page_model is not None:
            self._model_registry[page_model].add(func)

    def get_registered_info(self) -> Dict[str, dict]:
        info = defaultdict(dict)
        for k, v in self._template_registry.items():
            for f in v:
                key = f"{f.__module__}.{f.__name__}"
                info[key]["templates"] = info[key].get("templates", set())
                info[key]["templates"].add(k)
                info[key]["templates_count"] = len(info[key]["templates"])
        for k, v in self._model_registry.items():
            key = f"{f.__module__}.{f.__name__}"
            for f in v:
                info[key]["models"] = info[key].get("models", set())
                info[key]["models"].add(k)
                info[key]["models_count"] = len(info[key]["models"])
        return info

    def get_wrapper_context_func(
        self, template_path: str, page_model: Type["AbstractPage"]
    ) -> Callable:
        all_funcs = set()
        for cls in self._model_registry.keys():
            if issubclass(page_model, cls):
                all_funcs.update(self._model_registry[cls])
        all_funcs.update(self._template_registry[template_path])

        def context_func(request: HttpRequest, super_ctx: dict = {}):
            context = super_ctx.copy()
            for func in all_funcs:
                arg_spec = inspect.getfullargspec(func)
                kwargs = {}
                if "request" in arg_spec.args:
                    kwargs["request"] = request
                if "super_ctx" in arg_spec.args:
                    kwargs["super_ctx"] = context
                context.update(func(**kwargs) or {})
            return context

        return context_func

    def get_context_for_page(self, page: "AbstractPage", request, super_ctx: dict = {}):
        return self.get_wrapper_context_func(
            page.get_template_path(request), page.__class__
        )(request, super_ctx=super_ctx)


def register(
    template_path: Optional[str] = None,
    page_model: Optional[Type["AbstractPage"]] = None,
):
    assert (
        template_path is not None or page_model is not None
    ), "You must provide at least one between template_path and page_model"

    def inner(func: Callable):
        ctx_registry.register(func, template_path, page_model)
        return func

    return inner


ctx_registry = PagesContextRegistry()



import functools
from django.utils.translation import activate
from django.conf import settings


def active_lang(*args, **kwargs):
    def decorator(func):
        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            if len(args) and hasattr(args[0], "request"):
                request = args[0].request
            else:
                request = args[0] if len(args) else kwargs.get("request", None)
            lang = settings.LANGUAGE_CODE
            if request and hasattr(request, "GET"):
                lang = request.GET.get("lang", request.GET.get("language", lang))
            if request and hasattr(request, "data"):
                lang = request.data.pop("lang", request.data.pop("language", lang))
            if lang and lang in [lng[0] for lng in settings.LANGUAGES]:
                activate(lang)
                request.LANGUAGE_CODE = lang
            return func(*args, **kwargs)

        return wrapped_func

    return decorator



from django.utils.translation import gettext_lazy as _
from modeltranslation.utils import build_localized_fieldname
from rest_framework.exceptions import ValidationError

from camomilla.models.page import UrlNode
from camomilla.utils import activate_languages, is_page
from camomilla.utils.translation import get_nofallbacks, set_nofallbacks


class UniquePermalinkValidator:
    message = _("This slug generates a non-unique permalink.")

    requires_context = True

    def __call__(self, value, serializer):
        if not is_page(serializer.Meta.model):
            return
        errors = {}
        instance = serializer.instance
        exclude_kwargs = {}
        if instance and instance.url_node:
            exclude_kwargs["pk"] = instance.url_node.pk
        parent_page_field = getattr(
            getattr(serializer.Meta.model, "PageMeta", object),
            "parent_page_field",
            "parent_page",
        )
        parent_page = value.get(parent_page_field, None) or getattr(
            instance, parent_page_field, None
        )
        for language in activate_languages():
            f_name = build_localized_fieldname("slug", language)
            slug = value.get(f_name, instance and get_nofallbacks(instance, "slug"))
            fake_instance = serializer.Meta.model()
            set_nofallbacks(fake_instance, "slug", slug)
            if parent_page:
                set_nofallbacks(fake_instance, parent_page_field, parent_page)
            permalink = fake_instance.generate_permalink(safe=False)
            qs = UrlNode.objects.exclude(**exclude_kwargs)
            if qs.filter(permalink=permalink).exists():
                errors[f_name] = self.message
        if len(errors.keys()):
            raise ValidationError(errors)



from camomilla.models import Tag
from camomilla.permissions import CamomillaBasePermissions
from camomilla.serializers import TagSerializer
from camomilla.views.base import BaseModelViewset
from camomilla.views.mixins import BulkDeleteMixin, GetUserLanguageMixin


class TagViewSet(GetUserLanguageMixin, BulkDeleteMixin, BaseModelViewset):
    queryset = Tag.objects.all()
    serializer_class = TagSerializer
    permission_classes = (CamomillaBasePermissions,)
    model = Tag



from camomilla.models import Page
from camomilla.models.page import UrlNode
from camomilla.permissions import CamomillaBasePermissions
from camomilla.serializers import PageSerializer
from camomilla.serializers.page import UrlNodeSerializer
from camomilla.views.base import BaseModelViewset
from camomilla.views.decorators import active_lang
from camomilla.views.mixins import BulkDeleteMixin, GetUserLanguageMixin
from rest_framework.decorators import api_view
from rest_framework.response import Response
from django.shortcuts import get_object_or_404


class PageViewSet(GetUserLanguageMixin, BulkDeleteMixin, BaseModelViewset):
    queryset = Page.objects.all()
    serializer_class = PageSerializer
    permission_classes = (CamomillaBasePermissions,)
    model = Page


@active_lang()
@api_view(["GET"])
def fetch_page(request, permalink=""):
    node = get_object_or_404(UrlNode, permalink=f"/{permalink}")
    return Response(UrlNodeSerializer(node, context={"request": request}).data)



from django.db import models
from django.urls import reverse
from djsuperadmin.mixins import DjSuperAdminMixin


class AbstractContent(DjSuperAdminMixin, models.Model):
    identifier = models.TextField()
    content = models.TextField(default="")
    page = models.ForeignKey(
        "camomilla.Page",
        blank=False,
        null=True,
        on_delete=models.SET_NULL,
        related_name="contents",
    )

    @property
    def superadmin_get_url(self):
        return reverse("camomilla-content-djsuperadmin", kwargs={"pk": self.pk})

    @property
    def superadmin_patch_url(self):
        return reverse("camomilla-content-djsuperadmin", kwargs={"pk": self.pk})

    class Meta:
        abstract = True
        unique_together = ["identifier", "page"]

    def __str__(self):
        if len(self.identifier) > 40:
            return "%s..." % self.identifier[:40]
        return self.identifier


class Content(AbstractContent):
    pass



from rest_framework.response import Response
from django.db.models import Q
from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.contrib.postgres.search import SearchVector, SearchQuery, TrigramSimilarity


class TrigramSearchMixin:
    def handle_search(self, list_handler=None, search_fields=None):
        list_handler = list_handler if list_handler is not None else self.get_queryset()
        search_string = self.request.GET.get("search", None)
        search_fields = search_fields or getattr(self, "search_fields", [])
        if search_string and len(search_fields) > 0:
            filters = Q()
            for field in search_fields:
                filters |= Q(
                    **{f"search_{field}__gte": getattr(self, "trigram_threshold", 0.3)}
                )
            return list_handler.annotate(
                **{
                    f"search_{field}": TrigramSimilarity(field, search_string)
                    for field in search_fields
                },
            ).filter(filters)

        return list_handler


class PaginateStackMixin:
    def get_model(self, list_handler=None):
        list_handler = list_handler if list_handler is not None else self.get_queryset()
        return getattr(
            list_handler, "shared_model", getattr(list_handler, "model", None)
        )

    def parse_qs_value(self, string: str):
        if string and string.startswith("[") and string.endswith("]"):
            string = [self.parse_qs_value(substr) for substr in string[1:-1].split(",")]
        elif string and string.lower() in ["true", "false"]:
            string = string.lower() == "true"
        elif string and string.isdigit():
            string = int(string)
        return string

    def parse_filter(self, filter):
        filter_name, value = filter.split("=")
        return filter_name, self.parse_qs_value(value)

    def handle_pagination(self, list_handler=None, items_per_page=None):
        list_handler = list_handler if list_handler is not None else self.get_queryset()
        items_per_page = int(
            items_per_page
            or self.request.GET.get("items")
            or getattr(self, "items_per_page", 30)
        )
        paginator = Paginator(
            list_handler,
            items_per_page if items_per_page != -1 else list_handler.count(),
        )
        page = self.request.GET.get("page", 1) if items_per_page != -1 else 1

        try:
            elements = paginator.page(page)
        except PageNotAnInteger:
            # If page is not an integer, deliver first page.
            elements = paginator.page(1)
        except EmptyPage:
            # If page is out of range (e.g. 9999), deliver last page of results.
            elements = paginator.page(paginator.num_pages)

        return paginator, elements

    def handle_ordering(self, list_handler=None):
        list_handler = list_handler if list_handler is not None else self.get_queryset()
        sort = [p for p in self.request.GET.get("sort", "").split(",") if p]
        sort += list(self.get_model()._meta.ordering) + ["-pk"]
        order = self.request.GET.get("order", "asc")
        list_handler = list_handler.order_by(*sort)
        if order == "desc":
            list_handler = list_handler.reverse()
        return list_handler

    def handle_filters(self, list_handler=None):
        list_handler = list_handler if list_handler is not None else self.get_queryset()
        filters = dict(self.request.GET).get("fltr", [])
        for filter in filters:
            try:
                filter_name, value = self.parse_filter(filter)
                list_handler = list_handler.filter(**{filter_name: value})
            except Exception:
                pass
        return list_handler

    def handle_search(self, list_handler=None, search_fields=None):
        list_handler = list_handler if list_handler is not None else self.get_queryset()
        search_string = self.request.GET.get("search", None)
        search_fields = search_fields or getattr(self, "search_fields", [])
        if search_string and len(search_fields) > 0:
            return list_handler.annotate(
                search=SearchVector(*search_fields),
            ).filter(search=SearchQuery(search_string))

        return list_handler

    def handle_pagination_stack(
        self, list_handler=None, search_fields=None, items_per_page=None
    ):
        return self.handle_pagination(
            self.handle_ordering(
                self.handle_filters(self.handle_search(list_handler, search_fields))
            ),
            items_per_page,
        )

    def format_output(self, paginator, elements, SerializerClass=None):
        SerializerClass = SerializerClass or self.get_serializer_class()
        return {
            "items": SerializerClass(
                elements, many=True, context=self.get_serializer_context()
            ).data,
            "paginator": {
                "count": paginator.count,
                "page": elements.number,
                "has_next": elements.has_next(),
                "has_previous": elements.has_previous(),
                "pages": paginator.num_pages,
                "page_size": paginator.per_page,
            },
        }

    def list(self, *args, **kwargs):
        active = getattr(self, "force_active", False) or (
            self.request.GET.get("items", -1) != -1
        )
        if active:
            return Response(
                self.format_output(*self.handle_pagination_stack(self.get_queryset()))
            )
        else:
            return Response(
                self.get_serializer_class()(
                    self.handle_ordering(
                        self.handle_filters(self.handle_search(self.get_queryset()))
                    ),
                    many=True,
                    context=self.get_serializer_context(),
                ).data
            )



from camomilla.models import Article, Tag
from camomilla.serializers.base import BaseModelSerializer
from camomilla.serializers.mixins import AbstractPageMixin


class TagSerializer(BaseModelSerializer):
    class Meta:
        model = Tag
        fields = "__all__"


class ArticleSerializer(AbstractPageMixin, BaseModelSerializer):
    class Meta:
        model = Article
        fields = "__all__"



from django.conf import settings
from rest_framework.response import Response
from rest_framework import views


class LanguageViewSet(views.APIView):
    def get(self, request, *args, **kwargs):
        languages = []
        for key, language in settings.LANGUAGES:
            languages.append({"id": key, "name": language})
        return Response(
            {"language_code": settings.LANGUAGE_CODE, "languages": languages}
        )



from typing import Any, Callable, Union


def safe_getter(instance: Union[dict, object], key: str, default: Any = None) -> Any:
    if isinstance(instance, dict):
        return instance.get(key, default)
    return getattr(instance, key, default)


def pointed_getter(
    instance: Union[dict, object], pointed: str, default: Any = None
) -> Any:
    attrs = pointed.split(".", 1)
    data = safe_getter(instance, attrs[0], default)
    if len(attrs) == 2:
        data = pointed_getter(data, attrs[1], default)
    return data


def find_and_replace_dict(obj: dict, predicate: Callable):
    result = {}
    for k, v in obj.items():
        v = predicate(key=k, value=v)
        if isinstance(v, dict):
            v = find_and_replace_dict(v, predicate)
        result[k] = v
    return result



from __future__ import unicode_literals

from django.apps import AppConfig
from django.conf import settings


class CamomillaConfig(AppConfig):
    default_auto_field = "django.db.models.AutoField"
    name = "camomilla"

    def ready(self):
        migration_modules = getattr(settings, "MIGRATION_MODULES", {})
        if "camomilla" not in migration_modules:
            migration_modules["camomilla"] = "camomilla_migrations"
        setattr(settings, "MIGRATION_MODULES", migration_modules)
        self.module.autodiscover()



__version__ = "6.0.0-beta.12"


def get_core_apps():
    return ["rest_framework", "rest_framework.authtoken"]


def autodiscover():
    from camomilla.templates_context.autodiscover import autodiscover_context_files

    autodiscover_context_files()



from django.conf import settings as dj_settings
from django.db import models

from camomilla.models.page import AbstractPage
from camomilla import settings


class AbstractArticle(AbstractPage):
    content = models.TextField(default="")
    author = models.ForeignKey(
        dj_settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.SET_NULL
    )
    highlight_image = models.ForeignKey(
        "camomilla.Media",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="%(app_label)s_%(class)s_highlight_images",
    )
    tags = models.ManyToManyField("Tag", blank=True)

    class Meta:
        abstract = True
        ordering = ["ordering"]


class Article(AbstractArticle):
    class PageMeta:
        default_template = settings.ARTICLE_DEFAULT_TEMPLATE
        inject_context_func = settings.ARTICLE_INJECT_CONTEXT_FUNC


class AbstractTag(models.Model):
    name = models.CharField(max_length=200, unique=True)

    class Meta:
        abstract = True

    def __str__(self):
        return "(%s) %s" % (self.__class__.__name__, self.name)


class Tag(AbstractTag):
    pass



from enum import Enum
from uuid import uuid4
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.template.loader import render_to_string
from django.template import RequestContext
from django.utils.safestring import mark_safe
from pydantic import (
    Field,
    SerializationInfo,
    computed_field,
    model_serializer,
)
from camomilla import structured
from camomilla.models.page import UrlNode
from typing import Optional, Union, Callable, List


class LinkTypes(str, Enum):
    relational = "RE"
    static = "ST"


class MenuNodeLink(structured.BaseModel):
    link_type: LinkTypes = LinkTypes.static
    static: str = None
    content_type: int = None
    page_id: int = None
    url_node: UrlNode = None

    @model_serializer(mode="wrap", when_used="json")
    def update_relational(self, handler: Callable, info: SerializationInfo):
        if self.link_type == LinkTypes.relational:
            if self.content_type and self.page_id:
                c_type = ContentType.objects.filter(pk=self.content_type).first()
                model = c_type and c_type.model_class()
                page = model and model.objects.filter(pk=self.page_id).first()
                self.url_node = page and page.url_node
        return handler(self)

    def get_url(self, request=None):
        if self.link_type == LinkTypes.relational:
            return isinstance(self.url_node, UrlNode) and self.url_node.routerlink
        elif self.link_type == LinkTypes.static:
            return self.static

    @computed_field
    @property
    def url(self) -> Optional[str]:
        return self.get_url()


class MenuNode(structured.BaseModel):
    id: str = Field(default_factory=uuid4)
    meta: dict = {}
    nodes: List["MenuNode"] = []
    title: str = ""
    link: MenuNodeLink


class Menu(models.Model):
    key = models.CharField(max_length=200, unique=True, editable=False)
    available_classes = models.JSONField(default=dict, editable=False)
    enabled = models.BooleanField(default=True)
    nodes = structured.StructuredJSONField(default=list, schema=MenuNode)

    class Meta:
        verbose_name = _("menu")
        verbose_name_plural = _("menus")

    def render(
        self,
        template_path: str,
        request=None,
        context: Union[dict, RequestContext] = {},
    ):
        if isinstance(context, RequestContext):
            context = context.flatten()
        context.update({"menu": self})
        return mark_safe(render_to_string(template_path, context, request))

    class defaultdict(dict):
        def __missing__(self, key):
            dict.__setitem__(self, key, Menu.objects.get_or_create(key=key)[0])
            return self[key]

    def __str__(self) -> str:
        return self.key



from __future__ import unicode_literals
from collections import OrderedDict

from django.apps import AppConfig, apps
from django.conf import settings


def add_apps(*app_list):
    installed_apps = getattr(settings, "INSTALLED_APPS", [])
    changed = False
    for app in app_list:
        if app not in installed_apps:
            changed = True
            installed_apps = [app, *installed_apps]
    if changed:
        setattr(settings, "INSTALLED_APPS", installed_apps)
        apps.app_configs = OrderedDict()
        apps.apps_ready = apps.models_ready = apps.loading = apps.ready = False
        apps.clear_cache()
        apps.populate(settings.INSTALLED_APPS)


def set_default_settings(**kwargs):
    for key, value in kwargs.items():
        setattr(settings, key, getattr(settings, key, value))


class CamomillaThemeConfig(AppConfig):
    name = "camomilla.theme"

    def ready(self):
        set_default_settings(
            CKEDITOR_UPLOAD_PATH="editor-uploads/", X_FRAME_OPTIONS="SAMEORIGIN"
        )
        add_apps(
            "ckeditor_uploader",
            "ckeditor",
            "django_jsonform",
            "admin_interface",
            "colorfield",
        )



from rest_framework.decorators import action
from rest_framework.response import Response
from camomilla.permissions import CamomillaBasePermissions
from django.core.exceptions import ImproperlyConfigured
from django.db import transaction, router
from django.db.models.signals import post_save, pre_save
from itertools import chain
from django.db.models.expressions import F
from camomilla.fields import ORDERING_ACCEPTED_FIELDS


class OrderingMixin:
    @action(
        detail=False, methods=["post"], permission_classes=(CamomillaBasePermissions,)
    )
    def update_order(self, request):
        startId = int(request.data.get("startId"))
        endId = int(request.data.get("endId"))
        moved_items = list(self.move_item(startId, endId))
        return Response({"moved": moved_items, "reordered": self.reordered})

    @transaction.atomic
    def move_item(self, startId, endId):
        self.reordered = False
        model = self._get_model()
        (
            self.default_order_inverse,
            self.default_order_field,
        ) = self._get_default_ordering(model)
        rank_field = self.default_order_field

        startorder, endorder = self._get_order(startId, endId, model, rank_field)
        if endorder < startorder:
            move_filter = {
                f"{rank_field}__gte": endorder,
                f"{rank_field}__lte": startorder - 1,
            }
            move_delta = +1
            order_by = f"-{rank_field}"
        elif endorder > startorder:
            move_filter = {
                f"{rank_field}__gte": startorder + 1,
                f"{rank_field}__lte": endorder,
            }
            move_delta = -1
            order_by = rank_field
        else:
            return model.objects.none()

        obj_filters = {rank_field: startorder}

        try:
            obj = model.objects.get(**obj_filters)
        except (model.MultipleObjectsReturned, model.DoesNotExist):
            self._reorder(model)
            obj = model.objects.get(**obj_filters)

        move_qs = model.objects.filter(**move_filter).order_by(order_by)
        move_objs = list(move_qs)
        for instance in move_objs:
            setattr(instance, rank_field, getattr(instance, rank_field) + move_delta)
            pre_save.send(
                model,
                instance=instance,
                update_fields=[rank_field],
                raw=False,
                using=router.db_for_write(model, instance=instance),
            )
        move_qs.update(**{rank_field: F(rank_field) + move_delta})
        for instance in move_objs:
            post_save.send(
                model,
                instance=instance,
                update_fields=[rank_field],
                raw=False,
                using=router.db_for_write(model, instance=instance),
                created=False,
            )

        setattr(obj, rank_field, endorder)
        obj.save(update_fields=[rank_field])

        return [
            {"pk": instance.pk, "order": getattr(instance, rank_field)}
            for instance in chain(move_objs, [obj])
        ]

    def _get_default_ordering(self, model=None):
        model = model or self._get_model()
        try:
            _, prefix, field_name = model._meta.ordering[0].rpartition("-")
            model_field = model._meta.get_field(field_name)
            if not isinstance(model_field, ORDERING_ACCEPTED_FIELDS):
                raise ImproperlyConfigured(
                    f"Model {model.__module__}.{model.__name__} has a non-ordinable field '{model_field.__name__}' as first ordering key"
                )
        except (AttributeError, IndexError):
            raise ImproperlyConfigured(
                f"Model {model.__module__}.{model.__name__} requires a list or tuple 'ordering' in its Meta class"
            )
        return prefix == "-", field_name

    def _get_model(self):
        qs = self.get_queryset()
        return getattr(qs, "shared_model", getattr(qs, "model", None))

    def _get_order(self, startId, endId, model, rank_field):
        startorder = getattr(model.objects.get(pk=startId), rank_field)
        endorder = getattr(model.objects.get(pk=endId), rank_field)
        if startorder == endorder:
            self._reorder(model)
            return self._get_order(startId, endId, model, rank_field)
        return startorder, endorder

    def _reorder(self, model):
        self.reordered = True

        order = [
            "-" if self.default_order_inverse else "" + self.default_order_field,
            "pk" if self.default_order_inverse else "-pk",
        ]
        for index, obj in enumerate(model.objects.order_by(*order).iterator()):
            setattr(obj, self.default_order_field, index)
            obj.save()



from django.contrib.sitemaps import Sitemap
from camomilla.models import UrlNode


class CamomillaPagesSitemap(Sitemap):
    changefreq_default = "daily"
    priority_default = 0.5

    def changefreq(self, obj):
        if hasattr(obj.page.__class__, "changefreq"):
            return obj.page.changefreq
        return self.changefreq_default

    def priority(self, obj):
        if hasattr(obj.page.__class__, "priority"):
            return obj.priority
        return self.priority_default

    def items(self):
        return UrlNode.objects.filter(is_public=True)

    def lastmod(self, obj):
        if hasattr(obj.page.__class__, "lastmod"):
            return obj.page.lastmod
        return obj.date_updated_at


camomilla_sitemaps = {
    "pages": CamomillaPagesSitemap,
}



from django.shortcuts import redirect, render
from django.urls import path

from camomilla import settings
from django.conf import settings as django_settings
from .models import Page


def fetch(request, *args, **kwargs):
    preview = request.user.is_staff and request.GET.get("preview", False)
    append_slash = getattr(django_settings, "APPEND_SLASH", True)
    if append_slash and not request.path.endswith("/"):
        return redirect(request.path + "/")
    if "permalink" in kwargs:
        page = Page.get_or_404(
            request, bypass_public_check=preview, bypass_type_check=True
        )
    elif settings.AUTO_CREATE_HOMEPAGE is False:
        page, _ = Page.get_or_404(permalink="/", bypass_type_check=True)
    else:
        page, _ = Page.get_or_create_homepage()
    return render(request, page.get_template_path(request), page.get_context(request))


urlpatterns = [
    path("", fetch, name="camomilla-homepage"),
    path("<path:permalink>", fetch, name="camomilla-permalink"),
]
